// Copyright (c) 1999 arton

#include "stdafx.h"
#include "MacCapV2.h"
#include "Filter.h"
#include "MacPacket.h"

CFilter* CFilter::CreateFilter(IFilterParam* pParam)
{
	CFilter* pResult = NULL;
	VARIANT_BOOL fValid;
	if (pParam->get_Valid(&fValid) != S_OK || fValid == VARIANT_FALSE)
		return pResult;
	CFilter* pFilter[2];
	int cbFilter(0);
	FilterTypeConstants ftc;
	pParam->get_Type(&ftc);
	VARIANT var;
	VariantInit(&var);
	pParam->get_Value(&var);
	VARIANT_BOOL fSrc, fDst;
	pParam->get_Source(&fSrc);
	pParam->get_Destination(&fDst);
	USES_CONVERSION;
	switch (ftc)
	{
	case FT_MACADDRESS:
		SafeArrayLock(var.parray);
		if (fSrc)
		{
			CFilterValue* flt = new CFilterValue((LPBYTE)var.parray->pvData, 6, CMacPacket::PACKET_HEADER_SIZE + CMacPacket::SRC_MAC_ADDRESS);
			pFilter[cbFilter++] = flt;
		}
		if (fDst)
		{
			CFilterValue* flt = new CFilterValue((LPBYTE)var.parray->pvData, 6, CMacPacket::PACKET_HEADER_SIZE + CMacPacket::DST_MAC_ADDRESS);
			pFilter[cbFilter++] = flt;
		}
		SafeArrayUnlock(var.parray);
		break;
	case FT_IPADDRESS:
		if (fSrc)
		{
			CFilterValue* flt = new CFilterValue((LPBYTE)&var.lVal, sizeof(DWORD), CMacPacket::PACKET_HEADER_SIZE + CMacPacket::IPH_SRC, CFilter::IP_ARP);
			pFilter[cbFilter++] = flt;
		}
		if (fDst)
		{
			CFilterValue* flt = new CFilterValue((LPBYTE)&var.lVal, sizeof(DWORD), CMacPacket::PACKET_HEADER_SIZE + CMacPacket::IPH_DST, CFilter::IP_ARP);
			pFilter[cbFilter++] = flt;
		}
		break;
	case FT_ETHERTYPE:
		{
			WORD w = (WORD)htons(var.iVal);
			CFilterValue* flt = new CFilterValue((LPBYTE)&w, sizeof(WORD), CMacPacket::PACKET_HEADER_SIZE + CMacPacket::ETHER_TYPE);
			pFilter[cbFilter++] = flt;
		}
		break;
	case FT_PROTOCOL:
		{
			CFilterValue* flt = new CFilterValue((LPBYTE)&var.bVal, sizeof(BYTE), CMacPacket::PACKET_HEADER_SIZE + CMacPacket::IPH_PROTOCOL, CFilter::IP_ONLY);
			pFilter[cbFilter++] = flt;
		}
		break;
	case FT_TCPPORT:
	case FT_UDPPORT:
		if (fSrc)
		{
			WORD w = (WORD)htons(var.iVal);
			CFilterValue* flt = new CFilterValue((LPBYTE)&w, sizeof(WORD), CMacPacket::PACKET_HEADER_SIZE + CMacPacket::TCPH_SRCPORT, (ftc == FT_TCPPORT) ? CFilter::IP_TCP : CFilter::IP_UDP);
			pFilter[cbFilter++] = flt;
		}
		if (fDst)
		{
			WORD w = (WORD)htons(var.iVal);
			CFilterValue* flt = new CFilterValue((LPBYTE)&w, sizeof(WORD), CMacPacket::PACKET_HEADER_SIZE + CMacPacket::TCPH_DSTPORT, (ftc == FT_TCPPORT) ? CFilter::IP_TCP : CFilter::IP_UDP);
			pFilter[cbFilter++] = flt;
		}
		break;
	case FT_PACKETOFFSETVALUE:
		{
			SAFEARRAY* pAry = NULL;
			if (var.vt == (VT_ARRAY | VT_VARIANT))
			{
				pAry = var.parray;
			}
			else
			{
				pAry = *(var.pparray);
			}
			SafeArrayLock(pAry);
			VARIANT* pVar = reinterpret_cast<VARIANT*>(var.parray->pvData);
			int len = SysStringLen((pVar + 1)->bstrVal);
			LPBYTE pb = (LPBYTE)_alloca(len / 2 + 8);
			Pack(pb, (pVar + 1)->bstrVal, len);
			CFilterValue* flt = new CFilterValue(pb, ((len + 1) & 0xfffffffe) / 2, CMacPacket::PACKET_HEADER_SIZE + pVar->iVal);
			pFilter[cbFilter++] = flt;
			SafeArrayUnlock(pAry);
		}
		break;
	case FT_JOINT_AND:
	case FT_JOINT_OR:
		{
			SAFEARRAY* pAry = NULL;
			if (var.vt == (VT_ARRAY | VT_DISPATCH))
			{
				pAry = var.parray;
			}
			else
			{
				pAry = *(var.pparray);
			}
			if (ftc == FT_JOINT_AND)
				pFilter[cbFilter++] = new CFilterNodeManyAnd(pAry);
			else
				pFilter[cbFilter++] = new CFilterNodeManyOr(pAry);
		}
		break;
	case FT_NOT:
		{
			IFilterParam* pParam;
			if (var.pdispVal->QueryInterface(IID_IFilterParam, (void**)&pParam) == S_OK)
			{
				pFilter[cbFilter++] = new CFilterNot(pParam);
				pParam->Release();
			}
		}
		break;
	default:
		break;
	}
	VariantClear(&var);

	if (cbFilter == 1)
		pResult = pFilter[0];
	else if (cbFilter == 2)
		pResult = new CFilterNodeOr(pFilter[0], pFilter[1]);

	return pResult;
}
        
bool CFilterValue::DataCompare(int, LPBYTE pb) const
{
	switch (m_opt)
	{
	case IP_ONLY:
		if (*(WORD*)(pb + CMacPacket::PACKET_HEADER_SIZE + CMacPacket::ETHER_TYPE) != 0x0008)
			return false;
		break;
	case IP_ARP:
		if (*(WORD*)(pb + CMacPacket::PACKET_HEADER_SIZE + CMacPacket::ETHER_TYPE) != 0x0008
			&& *(WORD*)(pb + CMacPacket::PACKET_HEADER_SIZE + CMacPacket::ETHER_TYPE) != 0x0608)
			return false;
		break;
	case IP_TCP:
		if (*(WORD*)(pb + CMacPacket::PACKET_HEADER_SIZE + CMacPacket::ETHER_TYPE) != 0x0008
			|| *(pb + CMacPacket::PACKET_HEADER_SIZE + CMacPacket::IPH_PROTOCOL) != TCP)
			return false;
		break;
	case IP_UDP:
		if (*(WORD*)(pb + CMacPacket::PACKET_HEADER_SIZE + CMacPacket::ETHER_TYPE) != 0x0008
			|| *(pb + CMacPacket::PACKET_HEADER_SIZE + CMacPacket::IPH_PROTOCOL) != UDP)
			return false;
		break;
	default:
		break;
	}

	pb = pb + m_offset;
	if (m_cb <= sizeof(__int64))
	{
		switch (m_cb)
		{
		case sizeof(BYTE):
			return (*pb == (BYTE)m_hVal) ? true : false;
		case sizeof(WORD):
			return (*(WORD*)pb == (WORD)m_hVal) ? true : false;
		case sizeof(DWORD):
			return (*(DWORD*)pb == m_hVal) ? true : false;
		case sizeof(__int64):
			return (*(__int64*)pb == m_hVal) ? true : false;
		default:
			return (memcmp(&m_hVal, pb, m_cb) == 0) ? true : false;
		}
	}
	else
	{
		return (memcmp(m_pData, pb, m_cb) == 0) ? true : false;
	}
}

//////////////////////////////////////////////////////
// Helpers
LPBYTE Pack(LPVOID pv, LPCOLESTR lpc, int cb)
{
	LPBYTE pb = (LPBYTE)pv;
	LPBYTE pbsv = pb;
	*pb = 0;
	for (; cb; cb--, lpc++)
	{
		int c = towupper(*lpc);
		if (c >= L'0' && c <= L'9')
		{
			c -= L'0';
			if (cb & 1)
				*pb++ |= c;
			else
				*pb = c << 4;
		}
		else if (c >= L'A' && c <= L'F')
		{
			c -= L'A' - 10;
			if (cb & 1)
				*pb++ |= c;
			else
				*pb = c << 4;
		}
	}
	return pbsv;
}


HRESULT Unpack(VARIANT& v, BSTR* pResult)
{
	if (!pResult) return E_POINTER;
	VARIANT* pv;
	if (v.vt == (VT_BYREF | VT_VARIANT))
		pv = v.pvarVal;
	else
		pv = &v;
	SAFEARRAY* pAry = NULL;
	LPBYTE pb = NULL;
	int cb;
	if (pv->vt == (VT_ARRAY | VT_UI1))
	{
		pAry = pv->parray;
		SafeArrayLock(pAry);
		cb = pAry->rgsabound[0].cElements;
		pb = (LPBYTE)pAry->pvData;
	}
	else if (pv->vt == (VT_BYREF | VT_ARRAY | VT_UI1))
	{
		pAry = *pv->pparray;
		SafeArrayLock(pAry);
		cb = pAry->rgsabound[0].cElements;
		pb = (LPBYTE)pAry->pvData;
	}
	else if (pv->vt == VT_I4)
	{
		cb = sizeof(long);
		pb = (LPBYTE)&pv->lVal;
	}
	else if (pv->vt == VT_I2)
	{
		cb = sizeof(short);
		pb = (LPBYTE)&pv->iVal;
	}
	else
		return IDS_E_BADPARAM;

	LPOLESTR psz = (LPOLESTR)_alloca(cb * 2 + 16);
	*pResult = SysAllocString(Unpack(psz, pb, cb));
	if (pAry)
		SafeArrayUnlock(pAry);
	return S_OK;
}


LPOLESTR Unpack(LPOLESTR ps, LPVOID pv, int cb)
{
	LPOLESTR psv = ps;
	LPBYTE pb = (LPBYTE)pv;
	static const OLECHAR up[] = L"0123456789ABCDEF";
	for (; cb; cb--, pb++)
	{
		*ps++ = up[*pb >> 4];
		*ps++ = up[*pb & 0x0f];
	}
	*ps = L'\0';
	return psv;
}

