// Copyright(c) 1999 arton
//
// 1999 Jan. 15 (Waiting Event one by one)
//
#include "stdafx.h"
#include "MacCapV2.h"
#include "Adapters.h"
#include "MacPacket.h"
#include "MacCapture.h"
#include "PacketCollection.h"
#include "ProtocolNameCollection.h"
#include "EtherTypeNameCollection.h"
#include "Vendors.h"

/////////////////////////////////////////////////////////////////////////////
// CMacCapture
HRESULT STDMETHODCALLTYPE CMacCapture::get_MACAddress( 
            /* [retval][out] */ VARIANT __RPC_FAR *pRetVal)
{
	if (!m_pAdapters) return E_NOINTERFACE;
	if (m_hAdapter == INVALID_HANDLE_VALUE)
		return Error(IDS_E_NOTOPENED);
	if (!pRetVal) return E_POINTER;

	SAFEARRAYBOUND rgsabound[1];
	rgsabound[0].cElements = MacAddressSize;
	rgsabound[0].lLbound = 0;
	SAFEARRAY* pAry = SafeArrayCreate(VT_UI1, 1, rgsabound);
	SafeArrayLock(pAry);
	CopyMemory(pAry->pvData, m_MacAddress, MacAddressSize);
	SafeArrayUnlock(pAry);
	pRetVal->vt = VT_UI1 | VT_ARRAY;
	pRetVal->parray = pAry;
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CMacCapture::get_Adapters( 
            /* [retval][out] */ IAdapterCollection __RPC_FAR *__RPC_FAR *ppRetVal)
{
	if (!ppRetVal) return E_POINTER;
	if (!m_pAdapters) return E_NOINTERFACE;
	return m_pAdapters->QueryInterface(IID_IAdapterCollection, (void**)ppRetVal);
}
        
HRESULT STDMETHODCALLTYPE CMacCapture::Open( 
            /* [in] */ BSTR AdapterName)
{
	if (!m_pAdapters) return E_NOINTERFACE;
	if (m_hAdapter != INVALID_HANDLE_VALUE)
		CloseHandle();
	if (m_pStream)
	{
		m_pStream->Release();
		m_pStream = NULL;
	}

	USES_CONVERSION;
	LPTSTR pszAdapter;
	if (!AdapterName || !*AdapterName)
	{
		VarsIter it = (m_pAdapters->GetAdapters()).begin();
		pszAdapter = OLE2T((*it).bstrVal);
	}
	else
		pszAdapter = OLE2T(AdapterName);

	int cb = lstrlen(pszAdapter);
	LPTSTR psz = (LPTSTR)_alloca((cb + 8) * sizeof(TCHAR));
    wsprintf(psz, _T("\\\\.\\%s"), pszAdapter + 8);
    BOOL Result=DefineDosDevice(DDD_RAW_TARGET_PATH, &psz[4], pszAdapter);

    if (Result) 
	{
		m_hAdapter = CreateFile(psz, GENERIC_READ, 0, NULL, 
                             CREATE_ALWAYS,
                             FILE_FLAG_OVERLAPPED, NULL);

        if (m_hAdapter != INVALID_HANDLE_VALUE) 
		{
			m_szAdapter = new TCHAR[lstrlen(psz) + 8];
			lstrcpy(m_szAdapter, psz);
			DWORD dw;
			Result = DeviceIoControl(m_hAdapter, DIOC_CAPGetMacAddress,
				NULL, 0, m_MacAddress, MacAddressSize, &dw, NULL);
			ATLTRACE(_T("GetMac Result=%d, %d\n"), Result, dw);
            return S_OK;
        }
    }
	long l = GetLastError();
	ATLTRACE(_T("CreateFile Error = %d\n"), l);

	return Error(IDS_E_DRIVEROPEN, GUID_NULL, MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WIN32, l));
}
        
HRESULT STDMETHODCALLTYPE CMacCapture::Close( )
{
	if (!m_pAdapters) return S_OK;
	if (m_hAdapter != INVALID_HANDLE_VALUE)
	{
		Lock();
		m_fDuringCapture = false;
		Unlock();
		CloseHandle();
	}
	return S_OK;
}

HRESULT CMacCapture::CloseHandle()
{
	if (m_hThread)
	{
		DWORD dw;
		GetExitCodeThread(m_hThread, &dw);
		if (dw == STILL_ACTIVE)
		{
			PostThreadMessage(m_idThread, WM_QUIT, 0, 0);
			DWORD dw = GetTickCount() + 500;
			do
			{
				MSG msg;
				if (::PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
					::DispatchMessage(&msg);
				Sleep(100);
			}
			while (GetTickCount() < dw);
		}
		::CloseHandle(m_hThread);
		m_hThread = NULL;
	}
	if (m_hAdapter != INVALID_HANDLE_VALUE)
	{
		::CloseHandle(m_hAdapter);
		m_hAdapter = INVALID_HANDLE_VALUE;
	}
	if (m_szAdapter)
	{
		delete[] m_szAdapter;
		m_szAdapter = NULL;
	}
	ZeroMemory(m_MacAddress, MacAddressSize);
	return S_OK;
}

HRESULT STDMETHODCALLTYPE CMacCapture::Start(/* [in] */ VARIANT_BOOL SyncMode)
{
	if (!m_pAdapters) return E_NOINTERFACE;
	if (m_hAdapter == INVALID_HANDLE_VALUE)
		return Error(IDS_E_NOTOPENED);
	Lock();
	if (m_fDuringCapture)
	{
		Unlock();
		return Error(IDS_E_ALREADYSTART);
	}
	m_fDuringCapture = true;
	Unlock();
	if (m_pStream)
	{
		m_pStream->Release();
		m_pStream = NULL;
	}
	if (m_hThread == NULL)
	{
		IStream* ClientStream = NULL;
		if (m_vec.GetSize())
		{
			CComPtr<IUnknown> sp = m_vec.GetAt(0);
			CoMarshalInterThreadInterfaceInStream(IID_IDispatch, sp.p, &ClientStream);
		}
		ThreadParam* pParam = new ThreadParam;
		pParam->pParent = this;
		HANDLE hSem = pParam->hSem = CreateSemaphore(NULL, 0, 1, NULL);
		pParam->pClientSink = ClientStream;
		m_hThread = CreateThread(NULL, 0, RecvThreadProc, pParam, 0, &m_idThread);
		if (m_hThread == NULL)
		{
			delete pParam;
			m_fDuringCapture = false;
			return Error(IDS_E_THREAD);
		}
		else
		{
			SetThreadPriority(m_hThread, THREAD_PRIORITY_HIGHEST);
		}
		WaitForSingleObject(hSem, INFINITE);
		::CloseHandle(hSem);
	}
	HGLOBAL hMem = GlobalAlloc(GMEM_MOVEABLE | GMEM_SHARE, m_dwBuffSize);
	if (!hMem) return E_OUTOFMEMORY;
	if (CreateStreamOnHGlobal(hMem, TRUE, &m_pStream) != S_OK)
	{
		GlobalFree(hMem);
		m_pStream = NULL;
		m_fDuringCapture = false;
		return Error(IDS_E_STREAM);
	}
	m_dwPackets = m_dwWrote = m_dwLastWrote = 0;
	if (SyncMode)
		return SyncCapture(hMem);

	PostThreadMessage(m_idThread, WM_START_READ, (WPARAM)m_pStream, 0);
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CMacCapture::Stop( )
{
	if (!m_pAdapters) return E_NOINTERFACE;
	if (m_hAdapter == INVALID_HANDLE_VALUE)
		return Error(IDS_E_NOTOPENED);
	// for waiting for the I/O thread to finish to close the stream
	HANDLE hEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
	Lock();
	if (m_fDuringCapture)
	{
		m_fDuringCapture = false;
		if (m_hThread)
		{
			PostThreadMessage(m_idThread, WM_STOP_READ, (WPARAM)hEvent, 0);
		}
	}
	Unlock();
	WaitForSingleObject(hEvent, INFINITE);
	::CloseHandle(hEvent);
	return S_OK;
}

HRESULT STDMETHODCALLTYPE CMacCapture::put_BufferSize( 
            /* [in] */ long NewVal)
{
	if (NewVal <= 0) return Error(IDS_E_BADPARAM);
	m_dwBuffSize = NewVal;
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CMacCapture::get_BufferSize( 
            /* [retval][out] */ long __RPC_FAR *pRetVal)
{
	if (!pRetVal) return E_POINTER;
	*pRetVal = (long)m_dwBuffSize;
	return S_OK;
}

HRESULT STDMETHODCALLTYPE CMacCapture::put_RecvDataEnabled( 
            /* [in] */ VARIANT_BOOL NewVal)
{
	m_fRecvDataEnabled = (NewVal) ? true : false;
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CMacCapture::get_RecvDataEnabled( 
            /* [retval][out] */ VARIANT_BOOL __RPC_FAR *pRetVal)
{
	if (!pRetVal) return E_POINTER;
	*pRetVal = (m_fRecvDataEnabled) ? VARIANT_TRUE : VARIANT_FALSE;
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CMacCapture::put_MonitorEventInterval( 
            /* [in] */ long NewVal)
{
	m_dwMonitorInterval = NewVal;
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CMacCapture::get_MonitorEventInterval( 
            /* [retval][out] */ long __RPC_FAR *pRetVal)
{
	if (!pRetVal) return E_POINTER;
	*pRetVal = (long)m_dwMonitorInterval;
	return S_OK;
}

HRESULT STDMETHODCALLTYPE CMacCapture::SaveData( 
            /* [in] */ BSTR FileName,
            /* [retval][out] */ long __RPC_FAR *pResult)
{
	if (!pResult) return E_POINTER;
	if (!FileName || !*FileName) return Error(IDS_E_NOFILENAME);
	if (m_pStream == NULL) return Error(IDS_E_NODATA);
	if (m_fDuringCapture) return Error(IDS_E_ALREADYSTART);
	*pResult = 0;
	USES_CONVERSION;
	HANDLE h = CreateFile(OLE2T(FileName), GENERIC_WRITE, FILE_SHARE_READ, NULL,
		CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN, NULL);
	if (h == INVALID_HANDLE_VALUE)
		*pResult = GetLastError();
	else
	{
		STATSTG statstg;
		ZeroMemory(&statstg, sizeof(STATSTG));
		m_pStream->Stat(&statstg, STATFLAG_NONAME);
		HGLOBAL hg;
		HRESULT hr = GetHGlobalFromStream(m_pStream, &hg);
		if (hr != S_OK)
		{
			*pResult = (long)hr;
		}
		else
		{
			LPSTR p = (LPSTR)GlobalLock(hg);
			DWORD cb = statstg.cbSize.LowPart;
			DWORD dw;
			BOOL fResult;
			while (cb)
			{
				if (cb > 2048)
				{
					fResult = WriteFile(h, p, 2048, &dw, NULL);
					cb -= 2048;
					p += 2048;
				}
				else
				{
					fResult = WriteFile(h, p, cb, &dw, NULL);
					break;
				}
				if (fResult == FALSE)
				{
					*pResult = GetLastError();
					break;
				}
			}
			GlobalUnlock(hg);
		}
		::CloseHandle(h);
	}
	return S_OK;
}

HRESULT STDMETHODCALLTYPE CMacCapture::LoadData( 
            /* [in] */ BSTR FileName,
            /* [retval][out] */ long __RPC_FAR *pResult)
{
	if (!pResult) return E_POINTER;
	if (!FileName || !*FileName) return Error(IDS_E_NOFILENAME);
	if (m_fDuringCapture) return Error(IDS_E_ALREADYSTART);

	*pResult = 0;
	USES_CONVERSION;
	HANDLE h = CreateFile(OLE2T(FileName), GENERIC_READ, FILE_SHARE_READ, NULL,
		OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL);
	if (h == INVALID_HANDLE_VALUE)
		*pResult = GetLastError();
	else
	{
		if (m_pStream) 
			m_pStream->Release();

		if ((*pResult = (long)CreateStreamOnHGlobal(NULL, TRUE, &m_pStream)) == S_OK)
		{
			DWORD dwRead;
			do
			{
				char buff[2048];
				if (ReadFile(h, buff, 2048, &dwRead, NULL))
				{
					if (dwRead)
						m_pStream->Write(buff, dwRead, &dwRead);
				}
				else
				{
					*pResult = GetLastError();
					break;
				}
			}
			while (dwRead);
		}
		else
			m_pStream = NULL;
		::CloseHandle(h);
	}
	return S_OK;
}

HRESULT STDMETHODCALLTYPE CMacCapture::get_Packets( 
            /* [retval][out] */ IPacketCollection __RPC_FAR *__RPC_FAR *ppRetVal)
{
	if (!ppRetVal) return E_POINTER;
	if (!m_pStream) return Error(IDS_E_NODATA);
	if (m_fDuringCapture) return Error(IDS_E_ALREADYSTART);
	CComObject<CPacketCollection>* p;
	HRESULT hr = CComObject<CPacketCollection>::CreateInstance(&p);
	if (hr == S_OK)
	{
		p->InternalAddRef();
		p->SetStream(m_pStream);
	}
	*ppRetVal = p;
	return S_OK;
}

HRESULT STDMETHODCALLTYPE CMacCapture::RemoveFilters( void)
{
	if (m_fDuringCapture) return Error(IDS_E_ALREADYSTART);
	Lock();
	for (AFilterIter it = m_afilter.begin(); it != m_afilter.end(); it++)
		delete (*it);
	m_afilter.clear();
	Unlock();
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CMacCapture::AddFilter( 
            /* [in] */ IFilterParam __RPC_FAR *pParam)
{
	if (!pParam) return E_POINTER;
	if (m_fDuringCapture) return Error(IDS_E_ALREADYSTART);
	CFilter* pFilter = CFilter::CreateFilter(pParam);
	if (pFilter == NULL)
		return Error(IDS_E_BADPARAM);
	m_afilter.push_back(pFilter);
	return S_OK;
}

HRESULT STDMETHODCALLTYPE CMacCapture::get_Protocols( 
			/* [retval][out] */ IProtocolNameCollection __RPC_FAR *__RPC_FAR *ppRetVal)
{
	if (!ppRetVal) return E_POINTER;
	CComObject<CProtocolNameCollection>* p;
	HRESULT hr = CComObject<CProtocolNameCollection>::CreateInstance(&p);
	if (hr == S_OK)
	{
		p->InternalAddRef();
	}
	*ppRetVal = p;
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CMacCapture::get_EtherTypes( 
            /* [retval][out] */ IEtherTypeNameCollection __RPC_FAR *__RPC_FAR *ppRetVal)
{
	if (!ppRetVal) return E_POINTER;
	CComObject<CEtherTypeNameCollection>* p;
	HRESULT hr = CComObject<CEtherTypeNameCollection>::CreateInstance(&p);
	if (hr == S_OK)
	{
		p->InternalAddRef();
	}
	*ppRetVal = p;
	return S_OK;
}

HRESULT STDMETHODCALLTYPE CMacCapture::UnpackHelper( 
            /* [in] */ VARIANT v,
            /* [retval][out] */ BSTR __RPC_FAR *pResult)
{
	HRESULT hr = Unpack(v, pResult);
	if (hr != S_OK && HRESULT_FACILITY(hr) == FACILITY_NULL)
		return Error(HRESULT_CODE(hr));
	return hr;
}

HRESULT STDMETHODCALLTYPE CMacCapture::get_Vendors( 
            /* [retval][out] */ IVendorCollection __RPC_FAR *__RPC_FAR *ppRetVal)
{
	if (!ppRetVal) return E_POINTER;
	CComObject<CVendors>* p;
	HRESULT hr = CComObject<CVendors>::CreateInstance(&p);
	if (hr == S_OK)
	{
		p->InternalAddRef();
	}
	*ppRetVal = p;
	return S_OK;
}

////////////////////////////////////////////////////////////////////////
HRESULT CMacCapture::SyncCapture(HGLOBAL gh)
{
	LPBYTE pbStream = (LPBYTE)GlobalLock(gh);
	DWORD dwRead(0);
	DWORD dwPackets, dwWrote, dwTick, dwCurrentTick;
	dwPackets = dwWrote = 0;
	IStream* pStream;
	if (m_pStream->Clone(&pStream) != S_OK)
	{
		m_pStream->Release();
		m_fDuringCapture = false;
		return Error(IDS_E_STREAM);
	}
	HANDLE hEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
	HANDLE hStop = NULL;
	PostThreadMessage(m_idThread, WM_START_READ, (WPARAM)pStream, (LPARAM)hEvent);
	VARIANT_BOOL vStop;
	DWORD dwLastTick = GetTickCount();
	for (;;)
	{
		MSG msg;
		DWORD dw;
		dwTick = GetTickCount();
		if (dwTick < dwLastTick)
			dwLastTick = 0;
		dwTick -= dwLastTick;
		if (m_dwMonitorInterval)
		{
			if (dwTick >= m_dwMonitorInterval)
				dwTick = 10;
			else
				dwTick = m_dwMonitorInterval - dwTick;
		}
		else
			dwTick = 500;
		if (::PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
		{
			if (msg.message == WM_QUIT) break;
			TranslateMessage(&msg);
			DispatchMessage(&msg);
			continue;
		}
		else
		{
			dw = MsgWaitForMultipleObjects(1, &hEvent, FALSE, dwTick, QS_ALLINPUT);
			if (dw == WAIT_OBJECT_0 + 1) continue;
		}
		if (dw == WAIT_OBJECT_0)
		{
			if (m_fDuringCapture == false) break;	// Error occured

			while (dwRead <= m_dwLastWrote)	// m_dwLastWrote may not be volatile
			{
				CComObject<CMacPacket>* p;
				HRESULT hr = CComObject<CMacPacket>::CreateInstance(&p);
				if (hr == S_OK)
				{
					vStop = VARIANT_FALSE;
					m_pStream->AddRef();
					p->InternalAddRef();
					p->SetParam(m_pStream, dwRead);
					dwRead += *(DWORD*)(pbStream + dwRead);
					Fire_RecvData(p, &vStop);
					p->InternalRelease();
					if (vStop)
					{
						Lock();
						m_fDuringCapture = false;
						Unlock();
						hStop = CreateEvent(NULL, FALSE, FALSE, NULL);
						PostThreadMessage(m_idThread, WM_STOP_READ, (WPARAM)hStop, 0);
						break;
					}
				}
			}
		}
		else if (dw != WAIT_TIMEOUT)
		{	// error 
			break;
		}
		if (m_fDuringCapture == false)
			break;

		dwCurrentTick = GetTickCount();
		if (dwCurrentTick < dwLastTick) dwLastTick = 0;
		dwTick = dwCurrentTick - dwLastTick;
		if ((dwTick & 0xffffffff) >= (m_dwMonitorInterval & 0xffffffff))
		{
			if (m_dwMonitorInterval)
			{
				vStop = VARIANT_FALSE;
				Fire_MonitorEvent(dwTick, m_dwPackets - dwPackets, m_dwWrote - dwWrote, &vStop);
				if (vStop) 
				{
					Lock();
					m_fDuringCapture = false;
					Unlock();
					hStop = CreateEvent(NULL, FALSE, FALSE, NULL);
					PostThreadMessage(m_idThread, WM_STOP_READ, (WPARAM)hStop, 0);
					break;
				}
				else
				{
					dwPackets = m_dwPackets;
					dwWrote = m_dwWrote;
				}
			}
			dwLastTick = dwCurrentTick;
		}
	}
	if (hStop)
		WaitForSingleObject(hStop, INFINITE);
	else
		::Sleep(500);
	pStream->Release();
	::CloseHandle(hEvent);
	GlobalUnlock(gh);
	return S_OK;
}

///////////////////////////////////////////////////////////////////////////////////
const int MULTI_BUFF = 8;
unsigned long WINAPI CMacCapture::RecvThreadProc(  LPVOID lpParameter  )
{
	CoInitializeEx(NULL, COINIT_APARTMENTTHREADED | COINIT_DISABLE_OLE1DDE);

	ThreadParam* pParam = (ThreadParam*)lpParameter;
	CMacCapture* pParent = pParam->pParent;
	IDispatch* pNotifyDisp = NULL;
	bool fRecvNeed(false);
	int i;
	MSG msg;
	// Create Message Queue
	PeekMessage(&msg, NULL, 0, 0, PM_REMOVE);
	BYTE abBuffer[MULTI_BUFF][2048];
	bool fIO[MULTI_BUFF];
	int lastHandle(0);
	HANDLE hEvent[MULTI_BUFF], hParentEvent;
	OVERLAPPED ov[MULTI_BUFF];
	for (i = 0; i < MULTI_BUFF; i++)
	{
		memset(&ov[i], 0, sizeof(ov[0]));
		hEvent[i] = ov[i].hEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
		fIO[i] = false;
	}
	hParentEvent = NULL;
	if (pParam->pClientSink)
	{
		if (CoGetInterfaceAndReleaseStream(pParam->pClientSink, IID_IDispatch, (void**)&pNotifyDisp) != S_OK)
			pNotifyDisp = NULL;
	}
	ReleaseSemaphore(pParam->hSem, 1, NULL);
	delete pParam;
	IStream* pStream;
	ULARGE_INTEGER ulg;
	ulg.HighPart = 0;
	for (;;)
	{
		DWORD dwWait, dwRead[MULTI_BUFF], dwWritten, dwStop;
		BOOL Result;
		DWORD& dwWrote = ulg.LowPart;
		if (fRecvNeed)
		{
			for (i = 0; i < MULTI_BUFF; i++)
			{
				if (fIO[i] == false)
				{
					ResetEvent(hEvent[i]);
					Result = ReadFile(pParent->m_hAdapter, abBuffer[i], sizeof(abBuffer[0]), &dwRead[i], &ov[i]);
					long l = GetLastError();
					if (l != ERROR_IO_PENDING)
					{
						ATLTRACE(_T("ReadFile Error = %d\n"), GetLastError());
						for (int j = 0; j < MULTI_BUFF; j++) fIO[j] = false;
						CancelIo(pParent->m_hAdapter);
						lastHandle = 0;
						fRecvNeed = false;
						pStream->SetSize(ulg);
						StopCapture(pParent, hParentEvent, pNotifyDisp, l);
					}
					else
					{
						fIO[i] = true;
					}
				}
			}
			if (::PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
			{
				if (msg.message == WM_QUIT) break;
				dwWait = WAIT_OBJECT_0 + 1;
			}
			else
			{
				dwStop = FALSE;
				dwWait = MsgWaitForMultipleObjects(1, &hEvent[lastHandle], FALSE, INFINITE, QS_ALLINPUT);
				if (dwWait == WAIT_OBJECT_0 + 1)
				{
					continue;
				}
				else if (dwWait == WAIT_OBJECT_0)
				{
					i = lastHandle;
					lastHandle++;
					if (lastHandle >= MULTI_BUFF) lastHandle = 0;
					fIO[i] = false;
					Result = GetOverlappedResult(pParent->m_hAdapter, &ov[i], &dwRead[i], TRUE);
					if (Result == FALSE)
					{
						ATLTRACE(_T("Recv Failed Error=%d\n"), GetLastError());
					}
					else
					{
						ATLTRACE(_T("Recv-%d(%d) %dBytes\n"), i + 1, Result, dwRead[i]);
#ifdef _DEBUG
						for (int ii = 0; ii < 2; ii++)
						{
							for (int j = 0; j < 16; j++)
							{
								ATLTRACE(_T("%02X"), abBuffer[i][ii * 16 + j]);
							}
							ATLTRACE(_T("\n"));
						}
#endif
						// check filter
						if (pParent->m_afilter.empty() == false)
						{
							bool fSkip(true);
							for (AFilterIter it = pParent->m_afilter.begin(); it != pParent->m_afilter.end(); it++)
							{
								if ((*it)->CheckRange(dwRead[i]) && (*it)->DataCompare(dwRead[i], abBuffer[i]))
								{
									fSkip = false;
									break;
								}
							}
							if (fSkip) continue;
						}
						//
						*(DWORD*)abBuffer[i] = dwRead[i];
						dwWrote += dwRead[i];
						if (dwWrote > pParent->m_dwBuffSize)
						{
							ATLTRACE(_T("Buffer max \n"));
							dwWrote -= dwRead[i];	// for SetSize operation.
							dwStop = ERROR_BUFFER_OVERFLOW;
						}
						else
						{
							HRESULT hr = pStream->Write(abBuffer[i], dwRead[i], &dwWritten);
							if (hr != S_OK)
							{
								ATLTRACE(_T("Write Failure Result = %08X\n"), hr);
								dwStop = hr;
							}
							else
							{
								pParent->m_dwLastWrote = pParent->m_dwWrote;
								pParent->m_dwWrote = dwWrote;
								pParent->m_dwPackets++;
								if (pParent->m_fRecvDataEnabled && hParentEvent)
									PulseEvent(hParentEvent);
							}
						}
						if (dwStop)
						{
							CancelIo(pParent->m_hAdapter);
							lastHandle = 0;
							for (int j = 0; j < MULTI_BUFF; j++) fIO[j] = false;
							fRecvNeed = false;
							pStream->SetSize(ulg);
							StopCapture(pParent, hParentEvent, pNotifyDisp, dwStop);
						}
					}
					continue;
				}
				else	// Never. Because waiting with INFINITE.
				{
					continue;
				}
			}
		}
		else
		{
			if (GetMessage(&msg, 0, 0, 0) == FALSE)
				break;
		}
		// Message Procedure
		switch (msg.message)
		{
		case WM_START_READ:
			fRecvNeed = true;
			pStream = (LPSTREAM)msg.wParam;
			hParentEvent = (HANDLE)msg.lParam;
			dwWrote = 0;
			break;
		case WM_STOP_READ:
			if (fRecvNeed)
			{
				CancelIo(pParent->m_hAdapter);
				pStream->SetSize(ulg);
				lastHandle = 0;
			}
			for (i = 0; i < MULTI_BUFF; i++) fIO[i] = false;
			fRecvNeed = false;
			hParentEvent = NULL;
			if (msg.wParam)
			{
				SetEvent((HANDLE)msg.wParam);
			}
			break;
		default:
			DispatchMessage(&msg);
			break;
		}
	}
	if (fRecvNeed)
		CancelIo(pParent->m_hAdapter);
	if (pNotifyDisp)
		pNotifyDisp->Release();
	for (i = 0; i < MULTI_BUFF; i++)
		::CloseHandle(ov[i].hEvent);

	CoUninitialize();
	return 0;
}

void CMacCapture::StopCapture(CMacCapture* pParent, HANDLE& hParentEvent, IDispatch* pNotifyDisp, DWORD dwReason)
{
	CancelIo(pParent->m_hAdapter);
	if (pNotifyDisp)
	{
		CComVariant* pvars = new CComVariant[1];
		pvars[0] = (long)dwReason;
		DISPPARAMS disp = { pvars, NULL, 1, 0 };
		pNotifyDisp->Invoke(0x1, IID_NULL, LOCALE_USER_DEFAULT, DISPATCH_METHOD, &disp, NULL, NULL, NULL);
		delete[] pvars;
	}
	pParent->Lock();
	pParent->m_fDuringCapture = false;
	pParent->Unlock();
	if (hParentEvent)
	{
		SetEvent(hParentEvent);
		hParentEvent = NULL;
	}
}

///////////////////////////////////////////////////////////////////////////////////
STDMETHODIMP CMacCapture::InterfaceSupportsErrorInfo(REFIID riid)
{
	static const IID* arr[] = 
	{
		&IID_IMacCapture
	};
	for (int i=0; i < sizeof(arr) / sizeof(arr[0]); i++)
	{
		if (::InlineIsEqualGUID(*arr[i],riid))
			return S_OK;
	}
	return S_FALSE;
}

///////////////////////////////////////////////////////////////////////////////////
HRESULT CMacCapture::FinalConstruct()
{
	IAdapterCollectionEx* pAdapter;
	IClassFactory* pFac;
	HRESULT hr = _Module.GetClassObject(CLSID_AdapterCollectionEx, IID_IClassFactory, (void**)&pFac);
	if (hr != S_OK) return hr;
	hr = pFac->CreateInstance(NULL, IID_IAdapterCollectionEx, (void**)&pAdapter);
	pFac->Release();
	if (hr == S_OK)
	{
		m_pAdapters = static_cast<CAdapters*>(pAdapter);
	}

	return CoCreateFreeThreadedMarshaler(
		GetControllingUnknown(), &m_pUnkMarshaler.p);
}

void CMacCapture::FinalRelease()
{
	CloseHandle();
	if (m_pStream)
		m_pStream->Release();
	if (m_pAdapters)
		m_pAdapters->Release();
	m_pUnkMarshaler.Release();
}

HRESULT WINAPI CMacCapture::StreamAgg(void* pv, REFIID riid, LPVOID* ppv, DWORD dw)
{
	CMacCapture* pObj = static_cast<CMacCapture*>(pv);
	if (pObj->m_pStream)
	{
		IStream* pStream;
		HRESULT hr = pObj->m_pStream->Clone(&pStream);
		if (hr == S_OK)
		{
			LARGE_INTEGER dmove;
			dmove.QuadPart = 0;
			pStream->Seek(dmove, STREAM_SEEK_SET, NULL);
			*ppv = pStream;
		}
		return hr;
	}
	return E_NOINTERFACE;
}

